from .inference import data_inference, datafree_inference
from .inference import marginal_likelihood, posterior_predictive_sampling, process_log_marginal_likelihood
from .posterior_helper import dataplot, full_column, posterior_plot, flux, get_random_sample
from .splines import  spline_inference, get_spline_model, plot_spline_design, plot_spline
from .natural_spline import NaturalCubicSpline



import numpy as np
import specfit as sf
import matplotlib.pyplot as plt

# Data from J.E. Reynolds for J1939-6342
original_polynomial = [-30.7667,  26.4908, -7.0977, 0.605334]
original_nu0 = 1.0e6


sigma = 0.5

fig, ax = plt.subplots()
ax.set_xscale("log", nonpositive='clip')
ax.set_yscale("log", nonpositive='clip')
ax.grid(True);

nu = np.linspace(0.5e9, 10e9, 100)
S = sf.flux(nu, original_polynomial, nu0=original_nu0)
ax.plot(nu/1e9, S, label="polynomial fit")
ax.legend()
fig.tight_layout()
plt.show()

names, stats, a_cov, a_corr, f, fake_data = \
    sf.datafree_inference(
                        name="J1939-6342",  
                        freq_min=0.5e9, 
                        freq_max = 10e9, 
                        nfreq = 20, 
                        sigma = 0.5, 
                        a = original_polynomial,  
                        nu0 = original_nu0)

a = stats[0] # Means


print(f"Variables: {names}")
print(f"Means: {stats[0]}")
print(f"SDev: {stats[1]}")
print(f"Covariance Matrix:\n{np.array2string(a_cov, separator=',', precision=4)}")
print(f"Correlation Matrix:\n{np.array2string(a_corr, separator=',', precision=4)}")



import numpy as np
import matplotlib
matplotlib.use('PDF')

import matplotlib.pyplot as plt

from astropy.io import fits
from astropy import units as u
from astropy.coordinates import Angle

import specfit as sf

from piecewise_linear import piecewise_linear
import arviz as az


if __name__ == "__main__":
    with fits.open('Abell22_full.fits') as hdul:
        print(hdul.info())

        data = hdul[0].data
        h1 = hdul[0].header

        print(repr(h1))
        print(h1['COMMENT'])
        print(h1['SIMPLE'])
        print(h1['NTABLE'])
        print(data)

        data = hdul[1].data
        h1 = hdul[1].header
        print(repr(h1))

        ra = data['RA']
        dec = data['DEC']

        freq = []
        total_flux = []
        E_total_flux = []
        for i in range(12):
            freq.append(data[f"Frequency_{i}"])
            total_flux.append(data[f"Total_flux_{i}"])
            E_total_flux.append(data[f"E_Total_flux_{i}"])

        freq = np.array(freq).T
        total_flux = np.array(total_flux).T
        E_total_flux = np.array(E_total_flux).T
        print(f"RA shape {ra.shape}")
        print(f"DEC shape {dec.shape}")
        print(f"Freq Shape {freq.shape}")
        print(f"Flux Shape {total_flux.shape}")

        def vect2csv(a_list):
            converted_list = [str(element) for element in a_list] 
            return ", ".join(converted_list)

        def vect2hdr(name, dim):
            x = [f"{name}[{i}]" for i in range(dim)]
            return vect2csv(x)

        result_csv = []
        header = f"name, order, ra, dec, {vect2hdr("slopes",2)},  {vect2hdr("slopes_sigma",2)}, change_point, change_point_sigma, log_marginal_likelihood"
        result_csv.append(header)

        for i in range(len(ra)):

            r = ra[i]
            d = dec[i]
            nu = freq[i,:]
            S = total_flux[i,:]
            ES = E_total_flux[i,:]
            nu0 = 1e9
            sigma = ES

            def todms(a):
                angle = Angle(a * u.deg) 
                return angle.to_string(unit=u.degree)

            name=f"Source_RA:{todms(r)}_DEC:{todms(d)}"

            if False:
                x = np.log(nu/nu0)
                y = np.log(S)
                y0 = (S-ES/2)
                y1 = (S+ES/2)
                y_sigma = np.log(y1) - y
                x_curve = np.linspace(x[0], x[-1], 100)

                spline = sf.NaturalCubicSpline([x[0], x[-1]], n_interior_knots=1, degree=1)

                f = spline.get_expr()

                ret = spline.regression(x, y, y_sigma=y_sigma)
                print(ret)

                with plt.rc_context({"axes.grid": True, "axes.formatter.min_exponent": 2}):
                    fig, ax = plt.subplots(figsize=(6,4), layout='constrained')

                    ax.set_xscale("log", nonpositive='clip')
                    ax.set_yscale("log", nonpositive='clip')

                    ax.set_xlabel("Frequency (GHz)")
                    ax.set_ylabel("Flux (mJy)")

                    knot_y = ret["knot_y"]
                    y_best = f(k_0=ret["interior_knots"][0], x=x_curve, y_0=knot_y[0],
                            y_1=knot_y[1],
                            y_2=knot_y[2])

                    ax.plot(np.exp(x_curve), np.exp(y_best)*1000, label="fit")
                    ax.errorbar(np.exp(x), np.exp(y)*1000, yerr=ES, fmt='o', label="data")
                    ax.set_title(name)
                    ax.legend()
                    plt.savefig(f"{name}.pdf")
                # break
            else:
                idata, json_data = piecewise_linear(name, freq=nu, S=S, sigma=ES, nu0=nu0)
                json_data['ra'] = r
                json_data['dec'] = d

                line = f"{json_data['name']}, {json_data['order']}, {json_data['ra']}, {json_data['dec']}, {vect2csv(json_data['slopes'])}, {vect2csv(json_data['slopes_sigma'])}, {json_data['change_point']}, {json_data['change_point_sigma']}, {json_data['log_marginal_likelihood']}"
                result_csv.append(line)

                with open("results.csv", 'w') as csv_file:
                    for line in result_csv:
                        print(line, file=csv_file)

                del idata



#
# Inference Routines for the SpecFIT package.
#
# (c) 2023-2024 Tim Molteno tim@elec.ac.nz
#
import pymc as pm
import numpy as np
import arviz as az

import logging
import os
import traceback

from .posterior_helper import get_stats, chain_covariance, run_or_load
from .posterior_helper import Tflux as flux
from .posterior_helper import flux as num_flux

logger = logging.getLogger()




def get_model(name, freq, mu, sigma, order, nu0):

    mumax = np.max(mu)
    print(f"Mu max: {mumax}")
    print(f"Mu max log: {np.log(mumax)}")

    means = [0.0 for i in range(0, order)]
    sigmas = [2.5 for i in range(0, order)]
    means[0] = np.log(mumax)
    sigmas[0] = 2.5

    with pm.Model() as _model:
        _a = [pm.Normal(f"a[{i}]", mu=means[i], sigma=sigmas[i]) for i in range(order)]

        _x = pm.MutableData('frequencies', freq)  # a data container, mutable
        _brightness = flux(_x, _a, nu0)

        # Use a StudentT likelihood to deal with outliers 
        _likelihood = pm.StudentT("likelihood", nu=5, mu=_brightness,
                                  sigma=np.array(sigma), shape=_x.shape, observed=np.array(mu))
    return _model


def process_log_marginal_likelihood(smcmodel, n_samples):
    """
        Run the Sequential Monte Carlo sampling, and 
        process the annoying ragged function that is returned 
        by pymc
    """
    rng = np.random.default_rng(123)

    with smcmodel:
        smcidata = pm.sample_smc(draws=n_samples, threshold=0.6,
                            random_seed=rng,
                            chains=6)

    chain_lml = smcidata.sample_stats["log_marginal_likelihood"]

    # Process the ragged log_marginal_likelihood

    if chain_lml.sizes["chain"] != 6:
        chain_lml = chain_lml[0]
    
    lml_list = []
    for chain in chain_lml:
        chain_arr = np.array(chain).flatten()
        if (chain_arr.shape[0] == 1):
            chain_arr = chain_arr[0]
        last = chain_arr[-1]
        lml_list.append(last)
        
    lml = np.mean(lml_list)
    return lml


def marginal_likelihood(name, freq,
                        mu, sigma, nu0,
                        o_start=2, o_stop=6):
    """
    Estimate the marginal likelihood using Sequential Monte Carlo (SMC)
    sampling.

    The Marginal Likelihood represents the probability of the model itself,
    and can be used to choose the best order for a polynomial model.

    https://www.ma.imperial.ac.uk/~nkantas/sysid09_final_normal_format.pdf

    Parameters
    ----------
    name : str
        A unique key to identify this model. This is used to cache results
        of the inference.
    freq : array-like
        The frequencies at which  the measurements are made (Hz)
    mu : array-like
        The intensities (in Jansky) for each measurement
    sigma : array-like
        The estimate of the standard deviation of each measurement
    nu0 : float
        The normalizing frequency for the polynomial (typically 1 GHz)
    o_start : int
        The lowest order of polynomial fit.
    o_stop : int
        The hightes order of polynomial fit.

    Returns
    -------
    list
        a list of [order ml] where ml is the marginal likelihood
        (or Beyesian Evidence). This is scaled so that the maximum
        marginal likelihood is 1.0 (since we only know this up to
        a scaling)
    """

    ret = []
    rng = np.random.default_rng(123)

    for _ord in range(o_start, o_stop+1):
        try:
            _model = get_model(name, freq, mu, sigma, _ord, nu0=nu0)
            with _model:

                _evidence = process_log_marginal_likelihood(_model, 1500*_ord)
                # trace = pm.sample_smc(draws=1500*_ord, kernel=pm.smc.kernels.IMH,
                #                         chains=6, threshold=0.6,
                #                         correlation_threshold=0.01,
                #                         random_seed=rng, return_inferencedata=False, progressbar=False)
                # 
                # lml = trace.report.log_marginal_likelihood
                # _evidence = np.mean([chain[-1] for chain in lml])

                print(f"Log Marginal Likelihood: {_ord}:  {_evidence}")

                ret.append([_ord, _evidence])
        except Exception as e:
            print(f"Exception {e}")
            print(_idata.sample_stats)
            print(traceback.format_exc())
            ret.append([_ord, float("nan")])

    ret = np.array(ret)
    # bmax = np.max(ret[:,1])
    # ret[:,1] = np.exp(ret[:,1] - bmax)

    return ret


def posterior_predictive_sampling(idata, model, num_pp_samples):

    with model:
        pm.sample_posterior_predictive(idata, extend_inferencedata=True)

    # reduced_samples = az.extract(idata.posterior, num_samples=num_pp_samples)
    # post = reduced_samples.posterior
    post = az.extract(idata.posterior_predictive, num_samples=num_pp_samples)
    var_names = list(post.data_vars)

    ret = {}
    for v in var_names:
        ret[v] = post[v]

    return ret
    
    # az.plot_ppc(idata, num_pp_samples=100)


def data_inference(name, freq, mu, sigma, order, nu0,
                   n_samples=10000):
    """Infer a spectral polynomial from original measurements

    Use Bayesian inference to infer the coefficients of a polynomial model for
    the spectral data. 

    Parameters
    ----------
    name : str
        A unique key to identify this model. This is used to cache results
        of the inference.
    freq : array-like
        The frequencies at which  the measurements are made (Hz)
    mu : array-like
        The intensities (in Jansky) for each measurement
    sigma : array-like
        The estimate of the standard deviation of each measurement
    order : int
        The order of the polynomial fit.

    Returns
    -------
    list
        a list of strings representing the header columns
    """
    _model = get_model(name, freq, mu, sigma, order, nu0=nu0)
    _idata = run_or_load(_model, fname=f"idata_{name}.nc",
                         n_samples=n_samples, n_tune=n_samples,
                         cache=True)

    a_cov, a_corr, names = chain_covariance(_idata.posterior)
    stats, names = get_stats(_idata.posterior)

    return names, stats, a_cov, a_corr, _idata, _model


def datafree_inference(name, freq_min, freq_max, nfreq, sigma, a, nu0):
    """Infer a spectral polynomial covariance using data-free inference

    This is the further elaboration of the docstring. Within this section,
    you can elaborate further on details as appropriate for the situation.
    Notice that the summary and the elaboration is separated by a blank new
    line.

    Parameters
    ----------
    name : str
        A unique key to identify this model. This is used to cache results
        of the inference.
    freq : array-like
        The range of frequencies over which this inference should be made (Hz)
    freq : array-like
        The range of frequencies over which this inference should be made (Hz)
    """
    f = np.geomspace(freq_min, freq_max, nfreq)
    flux_data = (num_flux(f, np.array(a), nu0))
    rng = np.random.default_rng(12345)
    noise = rng.normal(loc=np.zeros_like(f), scale=sigma)

    ## Create fake data for the data-free inference step

    order = np.array(a).shape[0]
    # Now do the bayesian inference of the polynomial parameters.
    with pm.Model() as _model:
        fake_data = pm.ConstantData("y", flux_data, dims="obs_id")
        _a = [pm.Normal(f"a[{i}]", mu=a[i], sigma=2.5) for i in  range(order)]
        _brightness = flux(f, _a, nu0)
        _likelihood = pm.Normal("likelihood", mu=_brightness,
                                sigma=np.ones_like(f)*sigma,
                                observed=fake_data)
    _idata = run_or_load(_model, fname=f"idata_{name}.nc",
                         n_samples=5000, n_tune=order*3000)

    a_cov, a_corr, names = chain_covariance(_idata.posterior)
    stats, names = get_stats(_idata.posterior)

    return names, stats, a_cov, a_corr, f, flux_data



from pathlib import Path

import arviz as az
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pymc as pm
#import kmeans1d

# from patsy import dmatrix

from .posterior_helper import get_stats, chain_covariance,  spline_run_or_load
from .posterior_helper import Tflux as flux

# Need patsy and python3-graphviz 

RANDOM_SEED = 8927


def get_knots(x_values):

#    clusters, centroids = kmeans1d.cluster(x_values, 2)
#    print(clusters)
#    print(centroids)

    knot_list = np.linspace(x_values[0], x_values[-1], 1, endpoint=False)

    # knot_list = centroids
    return knot_list


def get_spline_design(x_values, y_values, knot_list):

    print(f"Knot List: {knot_list}")
    degree = 3
    spline_design = f"bs(freq, knots=knots, degree={degree}) - max"
    #spline_design = f"cr(freq, knots=knots) - max"
    return dmatrix(
        spline_design,
        {"freq": x_values,
         "knots": knot_list,
         "max": np.max(y_values)},
    )


def get_spline_model(name, freq, mu, sigma, nu0):
    # mumax = np.max(mu)
    # print(f"Mu max: {mumax}")
    # print(f"Mu max log: {np.log(mumax)}")
    # 
    # means = [0.0 for i in range(0, order)]
    # sigmas = [2.5 for i in range(0, order)]
    # means[0] = np.log(mumax)
    # sigmas[0] = 2.5
    # 
    # with pm.Model() as _model:
    #     _a = [pm.Normal(f"a[{i}]", mu=means[i], sigma=sigmas[i]) for i in range(order)]
    # 
    #     _x = pm.MutableData('frequencies', freq)  # a data container, mutable
    #     _brightness = flux(_x, _a, nu0)
    # 
    #     # Use a StudentT likelihood to deal with outliers 
    #     _likelihood = pm.StudentT("likelihood", nu=5, mu=_brightness,
    #                               sigma=np.array(sigma), shape=_x.shape, observed=np.array(mu))
    # 
    x_data = np.log(freq/nu0)
    y_data = np.log(mu)
    y_err = np.array(sigma)

    knot_list = get_knots(x_data)

    B = get_spline_design(x_data, y_data, knot_list)

    COORDS = {
                "splines": np.arange(B.shape[1]),
                "obs": range(len(y_data))
            }
    with pm.Model(coords=COORDS) as _model:
        # a = pm.Normal("a", 0, 3)
        w = pm.Normal("w", mu=0, sigma=2, size=B.shape[1], dims="splines")
        # mu = pm.Deterministic("mu", a + pm.math.dot(np.asarray(B, order="F"), w.T))
        # sigma = pm.Exponential("sigma", 1)
        mu = pm.Deterministic("mu",  pm.math.dot(np.asarray(B, order="F"), w.T))
        D = pm.Normal("likelihood", mu=mu, sigma=y_err, observed=y_data, dims="obs")

    return _model, knot_list


def spline_inference(name, freq, mu, sigma, nu0,
                   n_samples=1000):
    """Infer a spectral spline from original measurements

    Use Bayesian inference to infer the coefficients of a polynomial model for
    the spectral data. 

    Parameters
    ----------
    name : str
        A unique key to identify this model. This is used to cache results
        of the inference.
    freq : array-like
        The frequencies at which  the measurements are made (Hz)
    mu : array-like
        The intensities (in Jansky) for each measurement
    sigma : array-like
        The estimate of the standard deviation of each measurement
    order : int
        The order of the polynomial fit.

    Returns
    -------
    list
        a list of strings representing the header columns
    """
    _model, knot_list = get_spline_model(name, freq, mu, sigma, nu0=nu0)
    _idata = spline_run_or_load(_model, fname=f"idata_{name}_{len(knot_list)}.nc",
                                n_samples=n_samples, n_tune=n_samples,
                                cache=False)

    print(_idata.posterior.keys())

    chain = 0
    a = _idata.posterior.get('w')[chain, :]
    a = np.array(a)
    a_cov, a_corr = np.cov(a), np.corrcoef(a)

    stats, names = get_stats(_idata.posterior)

    return names, stats, a_cov, a_corr, _idata, _model, knot_list


# x_data = np.log(freq/nu0)
# y_data = np.log(S)
# y_err = 0.01
# 
# 
# num_knots = 3
# 
# clusters, centroids = kmeans1d.cluster(x_data, num_knots)
# print(clusters)
# print(centroids)
# 
# knot_list = centroids
# # knot_list = np.linspace(x_data[0], x_data[-1], num_knots, endpoint=True)
# print(f"Knot List: {knot_list}")
# 

def plot_spline_design(freq, mu, nu0, knot_list):
    x_data = np.log(freq/nu0)
    y_data = np.log(mu)
    
    test_freq = np.linspace(x_data[0], x_data[-1], 100)
    B = get_spline_design(test_freq, y_data, knot_list)
    spline_df = (
        pd.DataFrame(B)
        .assign(freq=test_freq)
        .melt("freq", var_name="spline_i", value_name="value")
    )

    color = plt.cm.magma(np.linspace(0, 0.80, len(spline_df.spline_i.unique())))

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)

    for i, c in enumerate(color):
        subset = spline_df.query(f"spline_i == {i}")
        subset.plot("freq", "value", c=c, ax=ax, label=i)
    ax.legend(title="Spline Index", loc="upper center", fontsize=8, ncol=6);
    plt.show()


def plot_spline(idata, freq, mu, nu0, knot_list):

    x_data = np.log(freq/nu0)
    y_data = np.log(mu)

    test_freq = np.linspace(x_data[0], x_data[-1], 100)

    wp = idata.posterior["w"].mean(("chain", "draw")).values

    B = get_spline_design(test_freq, y_data, knot_list)

    spline_df = (
        pd.DataFrame(B * wp.T)
        .assign(freq=test_freq)
        .melt("freq", var_name="spline_i", value_name="value")
    )

    spline_df_merged = (
        pd.DataFrame(np.dot(B, wp.T))
        .assign(freq=test_freq)
        .melt("freq", var_name="spline_i", value_name="value")
    )

    color = plt.cm.rainbow(np.linspace(0, 1, len(spline_df.spline_i.unique())))

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)

    for i, c in enumerate(color):
        subset = spline_df.query(f"spline_i == {i}")
        subset.plot("freq", "value", c=c, ax=ax, label=i)
    spline_df_merged.plot("freq", "value", c="black", lw=2, ax=ax)
    ax.legend(title="Spline Index", loc="lower center", fontsize=8, ncol=6)

    ax.plot(x_data, y_data, 'o')
    for knot in knot_list:
        ax.axvline(knot, color="grey", alpha=0.4);

    plt.show()

# B = get_spline_design(x_data)
# 
# COORDS = {
#             "splines": np.arange(B.shape[1]),
#             "obs": range(len(y_data))
#          }
# with pm.Model(coords=COORDS) as spline_model:
#     # a = pm.Normal("a", 0, 3)
#     w = pm.Normal("w", mu=0, sigma=2, size=B.shape[1], dims="splines")
#     # mu = pm.Deterministic("mu", a + pm.math.dot(np.asarray(B, order="F"), w.T))
#     # sigma = pm.Exponential("sigma", 1)
#     mu = pm.Deterministic("mu",  pm.math.dot(np.asarray(B, order="F"), w.T))
#     D = pm.Normal("D", mu=mu, sigma=y_err, observed=y_data, dims="obs")


# pm.model_to_graphviz(spline_model)
# plt.show()
#
# 
# 
# original_data = np.array(
#     [[0.408,  6.24, 0.312 ],
#      [0.843, 13.65, 0.6825],
#      [1.38 , 14.96, 0.748 ],
#      [1.413, 14.87, 0.7435],
#      [1.612, 14.47, 0.7235],
#      [1.66 , 14.06, 0.703 ],
#      [1.665, 14.21, 0.7105],
#      [2.295, 11.95, 0.5975],
#      [2.378, 11.75, 0.5875],
#      [4.8  ,  5.81, 0.2905],
#      [4.8  ,  5.76, 0.288 ],
#      [4.835,  5.72, 0.286 ],
#      [4.85 ,  5.74, 0.287 ],
#      [8.415,  2.99, 0.1495],
#      [8.42 ,  2.97, 0.1485],
#      [8.64 ,  2.81, 0.1405]])
# 
# freq_ghz, S, delta_S = original_data.T
# freq = freq_ghz*1e9
# nu0=1.0e9
# 
# order = 3
# 
# name = "j1939-6342"
# mu = S
# sigma = delta_S
# 
# spline_model = get_spline_model(name, freq, mu, sigma, order, nu0=nu0)
    # _idata = run_or_load(_model, fname=f"idata_{name}.nc",
    #                      n_samples=n_samples, n_tune=n_samples)

# with spline_model:
#     idata = pm.sample_prior_predictive()
#     idata.extend(pm.sample(draws=1000, tune=3000, random_seed=RANDOM_SEED, chains=4))
#     pm.sample_posterior_predictive(idata, extend_inferencedata=True)
# 
# print(idata.keys())
# print(az.summary(idata, var_names=[ "w"]))
# 
# az.plot_trace(idata, var_names=[ "w"]);
# plt.show()
# 
# ## Pair Plots
# 
# az.plot_pair(
#         idata,
#         var_names=['w'],
#         kind="hexbin",
#         filter_vars="like",
#         marginals=True,
#         figsize=(12, 12),
#     )
# plt.tight_layout()
# #plt.savefig(f"posterior_pairs.pdf")
# plt.show()
# 
# #  Do some posterior sampling
# 
# wp = idata.posterior["w"].mean(("chain", "draw")).values
# 
# B = get_spline_design(test_freq)
# 
# spline_df = (
#     pd.DataFrame(B * wp.T)
#     .assign(freq=test_freq)
#     .melt("freq", var_name="spline_i", value_name="value")
# )
# 
# spline_df_merged = (
#     pd.DataFrame(np.dot(B, wp.T))
#     .assign(freq=test_freq)
#     .melt("freq", var_name="spline_i", value_name="value")
# )
# 
# 
# color = plt.cm.rainbow(np.linspace(0, 1, len(spline_df.spline_i.unique())))
# fig = plt.figure()
# for i, c in enumerate(color):
#     subset = spline_df.query(f"spline_i == {i}")
#     subset.plot("freq", "value", c=c, ax=plt.gca(), label=i)
# spline_df_merged.plot("freq", "value", c="black", lw=2, ax=plt.gca())
# plt.legend(title="Spline Index", loc="lower center", fontsize=8, ncol=6)
# 
# plt.plot(x_data, y_data, 'o')
# for knot in knot_list:
#     plt.gca().axvline(knot, color="grey", alpha=0.4);
# 
# plt.show()



import numpy as np
import specfit as sf
import matplotlib.pyplot as plt
import sympy as sp

# Data from J.E. Reynolds for J1939-6342
original_data = np.array(
    [[0.408,  6.24, 0.312 ],
     [0.843, 13.65, 0.6825],
     [1.38 , 14.96, 0.748 ],
     [1.413, 14.87, 0.7435],
     [1.612, 14.47, 0.7235],
     [1.66 , 14.06, 0.703 ],
     [1.665, 14.21, 0.7105],
     [2.295, 11.95, 0.5975],
     [2.378, 11.75, 0.5875],
     [4.8  ,  5.81, 0.2905],
     [4.8  ,  5.76, 0.288 ],
     [4.835,  5.72, 0.286 ],
     [4.85 ,  5.74, 0.287 ],
     [8.415,  2.99, 0.1495],
     [8.42 ,  2.97, 0.1485],
     [8.64 ,  2.81, 0.1405],
     [8.64 ,  2.81, 0.1405]])

freq_ghz, mu, sigma = original_data.T
freq = freq_ghz*1e9
nu0=1.0e9

if True:
    names, stats, a_cov, a_corr, idata, model = \
        sf.data_inference("J1939-6342",
            freq=freq, mu=mu, sigma=sigma,
            order=4, nu0=nu0, n_samples=5000)

    a = stats[0] # Means

    fig, ax = sf.dataplot(plt, "J1939-6342", freq=freq, mu=mu, sigma=sigma)
    nu = np.linspace(freq[0], freq[-1], 100)
    S = sf.flux(nu, a, nu0=nu0)
    ax.plot(nu/1e9, S, label="polynomial fit")
    ax.legend()
    fig.tight_layout()
    plt.show()

    print(f"Variables: {names}")
    print(f"Means: {stats[0]}")
    print(f"SDev: {stats[1]}")
    print(f"Covariance Matrix:\n{np.array2string(a_cov, separator=',', precision=4)}")
    print(f"Correlation Matrix:\n{np.array2string(a_corr, separator=',', precision=4)}")

    post = idata.posterior
    print(post.keys())

    pp = sf.posterior_predictive_sampling(idata, model, 1000)
    print(pp.keys())

    x = sp.Symbol('x', real=True)
    polys = None
    for k in pp.keys():
        n = int(k[2])
        a = pp[k].values

        print(a.shape)
        n_samples = len(a)
        if polys is None:
            polys = [0]*n_samples

        for i in range(n_samples):
            polys[i] += a[i]*x**n

    peaks = []
    for p in polys:
        eqn = sp.Eq(sp.diff(p, x), 0)
        low, high = sp.solveset(eqn, x, domain=sp.S.Reals)

        f = nu0*sp.exp(low)
        print(f"{eqn} --> {low} --> {f}")
        peaks.append(sp.N(f))
        # logw = log(nu/nu0) = soln
        # --> nu = nu0*exp(soln)

    peaks = np.array(peaks, dtype=float)/1e9
    print(peaks.shape)
    plt.hist(peaks, bins='fd')
    plt.title("PDF of spectral peaks J1939-6342")
    plt.xlabel("Frequency (GHz)")
    plt.grid(True)
    plt.savefig("j1939_peaks.pdf")
    plt.show()
## Now do polynomial inference from the data again.

names2, stats2, a_cov2, a_corr2, f, fake_data = sf.datafree_inference('J1939-6342-poly', freq_min=freq[0], freq_max=freq[-1], nfreq=20, sigma=0.5, a=[ 2.69445071,  0.24791307, -0.71448904,  0.11324043], nu0=nu0)

print(fake_data)

fig, ax = plt.subplots()
a = stats2[0] # Means
ax.set_xscale("log", nonpositive='clip')
ax.set_yscale("log", nonpositive='clip')
ax.errorbar(f/1e9, fake_data, yerr=0.5, fmt='.', label="Fake Data")

ax.set_xlabel("Frequency (GHz)")
ax.set_ylabel("Flux (Jy)")
ax.grid(True)

nu = np.linspace(f[0], f[-1], 100)
S = sf.flux(nu, a, nu0=nu0)
ax.plot(nu/1e9, S, label="data-free polynomial fit")
ax.legend()
fig.tight_layout()
plt.show()

print(f"Variables: {names2}")
print(f"Means: {stats2[0]}")
print(f"SDev: {stats2[1]}")
print(f"Covariance Matrix:\n{np.array2string(a_cov2, separator=',', precision=4)}")
print(f"Correlation Matrix:\n{np.array2string(a_corr2, separator=',', precision=4)}")






import sympy as sp
import matplotlib.pyplot as plt 
import numpy as np
import scipy

'''
    Natural Spline Fitting
    
    Author Tim Molteno tim@elec.ac.nz (c) 2024
    
    This code performs a spline regression to fit a spline to some data. The spline is natural 
    in that the second derivative on the boundary is zero. The position of the interior knots is
    a parameter in this code. 

'''
## Changepoint pymc https://ckrapu.github.io/blog/2022/nonparametric-changepoint-model-pymc/
### https://causalpy.readthedocs.io/en/latest/notebooks/rkink_pymc.html

def poly(x, x_0, x_1, n, degree=3):
    #
    # Create a polynomial that will be bounded between x_0, and x_1 which is the n_th spline interval.
    #
    a = [sp.Symbol(f"a_{n}_{i}") for i in range(degree+1)]

    _poly = 0
    for i in range(degree+1):
        _poly += a[i]*x**i
    return _poly, a


def window_below(x, x_1):
    return (1 - sp.Heaviside(x - x_1))


def window_above(x, x_0):
    return sp.Heaviside(x - x_0)


def window(x, x_0, x_1):
    return sp.Heaviside(x - x_0) * (1 - sp.Heaviside(x - x_1))


def natural_spline(x, knot_x, knot_y, degree):

    # The sum of piecewise step functions each that contains a cubic polynomial. We don't include the boundary knots

    n_knots = len(knot_x)
    n_intervals = n_knots - 1
    polys = []
    variables = []
    windows = []
    for i in range(n_intervals):
        x_0 = knot_x[i]
        x_1 = knot_x[i+1]
        print(f"interval {i} : {x_0},{x_1}")
        p, pvars = poly(x, x_0, x_1, i, degree)
        polys.append(p)
        variables.extend(pvars)
        if i == 0:
            windows.append(window_below(x, x_1))
        elif i == n_intervals-1:
            windows.append(window_above(x, x_0))
        else:
            windows.append(window(x, x_0, x_1))

    eqns = []
    for i in range(1, n_knots-1):
        x_0 = knot_x[i]
        y_0 = knot_y[i]
        print(f"interior knots {i} : {x_0}")
        p_0 = polys[i-1]
        p_1 = polys[i]

        # Set interior values
        eqns.append(sp.Eq(p_0.subs(x, x_0), y_0))
        eqns.append(sp.Eq(p_1.subs(x, x_0), y_0))
        # Match interior points
        eqns.append(sp.Eq(p_0.subs(x, x_0), p_1.subs(x, x_0)))
        # Match interior derivatives
        for d in range(degree):
            eqns.append(sp.Eq(sp.diff(p_0, x, d).subs(x, x_0),
                              sp.diff(p_1, x, d).subs(x, x_0)))

    for i in [0, -1]:  # Boundary knots
        # Second derivatives to zero at the boundary
        eqns.append(sp.Eq(sp.diff(polys[i], x, 2).subs(x, knot_x[i]), 0))
        # Set exterior points to their values
        eqns.append(sp.Eq(polys[i].subs(x, knot_x[i]), knot_y[i]))

    soln = sp.solve(eqns, variables)
    reduced_polys = []
    for p in polys:
        rp = sp.simplify(p.subs(soln))
        reduced_polys.append(rp)

    spline = 0
    for p, w in zip(reduced_polys, windows):
        spline = spline + w * p

    expr = sp.simplify(spline.subs(soln))
    free_syms = sorted(expr.free_symbols, key = lambda symbol: symbol.name)

    return expr, eqns, free_syms, reduced_polys


class NaturalCubicSpline:
    #
    #   A natural spline of degree n that can be evaluated using numpy, with unknown knot positions
    #   TODO: The case degree 2 does not work
    #
    def __init__(self, interval, n_interior_knots, degree=3):

        x = sp.Symbol('x')

        x_0, x_1 = interval

        k = [sp.Symbol(f"k_{n}") for n in range(n_interior_knots)]
        all_knots = [[x_0], k, [x_1]]
        print(all_knots)

        knot_x = [item for row in all_knots for item in row]
        knot_y = [sp.Symbol(f"y_{n}") for n in range(len(knot_x))]

        self.expr, self.eqns, self.variables, self.polys = \
            natural_spline(x, knot_x=knot_x,
                           knot_y=knot_y,
                           degree=degree)
        print(self.variables)
        print(self.eqns)
        print(self.expr)
        for p in self.polys:
            print(p)
        self.n_interior_knots = n_interior_knots
        self.x_0, self.x_1 = interval
        self.degree = degree
        self.n_param = len(self.variables) - 1

    def get_expr(self):
        f = sp.lambdify(self.variables, self.expr, cse=True)
        return f

    def regression(self, x_data, y_data, y_sigma):
        # Calculate the best fit spline to match the data in a least squares sense
        expr = self.get_expr()
        # The arguments of expr, are in self.variables (in order), these are {k_i}, x, {y_i}
        # The parameters are {k_i}{y_i}

        def f(v):
            param = v[0:self.n_interior_knots].tolist()
            param.append(x_data)
            param.extend(v[self.n_interior_knots:].tolist())
            residual = (expr(*param) - y_data)/y_sigma  
            # This is weighted so that the residual is weighted by 1/sigma^2
            # see https://en.wikipedia.org/wiki/Weighted_least_squares
            return np.sum(residual**2)


        bounds = []
        v0 = []
        for i in range(1, self.n_interior_knots+1):
            bounds.append((self.x_0, self.x_1))
            v0.append(self.x_0 + i*(self.x_1 - self.x_0)/(self.n_interior_knots+1))

        n_knots = self.n_param - self.n_interior_knots
        for i in range(n_knots):
            bounds.append((None, None))
            ind = int(i*y_data.shape[0]/n_knots)
            print(ind)
            v0.append(y_data[ind])
        print(f"v0 = {v0}, var={self.variables}")

        v0 = np.array(v0)
        f0 = f(v0)
        print(f0)

        opt = scipy.optimize.minimize(f, v0, method="Nelder-Mead",
                                      bounds=bounds)
        print(opt)
        v_best = opt.x

        ret = {}
        ret["interior_knots"] = v_best[0:self.n_interior_knots].tolist()
        ret["interval"] = [self.x_0, self.x_1]
        ret["degree"] = self.degree

        k_x = []
        k_x.append(self.x_0)
        k_x.extend(v_best[0:self.n_interior_knots].tolist())
        k_x.append(self.x_1)

        ret["knot_x"] = k_x
        ret["knot_y"] = v_best[self.n_interior_knots:].tolist()

        return ret


if __name__ == "__main__":
    spline = NaturalCubicSpline([-1, 3], n_interior_knots=1, degree=3)

    f = spline.get_expr()

    x = np.linspace(-1,3,20)
    y = np.sin(x)+np.random.uniform(size=(20,))

    ret = spline.regression(x, y, y_sigma=0.1)
    print(ret)

    plt.plot(x, y, '.', label="data")

    knot_y = ret["knot_y"]
    y_best = f(k_0=ret["interior_knots"][0], x=x, y_0=knot_y[0],
               y_1=knot_y[1],
               y_2=knot_y[2])

    plt.plot(x, y_best, label="spline regression")
    plt.legend()
    # plt.plot(2.5, 2, 'o')
    plt.show()



#!/usr/bin/env python
# coding: utf-8

# # Calculating Priors to be used for Calibration Source Spectra
# 
# Tim Molteno tim@elec.ac.nz
# 
# A mini project on calculating the uncertainties in the source spectral parameters. The Perley and Butler catalogs are typically 0.5 - 1.5% accurate (Oleg, Personal Communication). What I'd like to do for inference is decide on a prior for the spectral parameters $I_0$ and $\{a_i\}$.
# 
# 
# ## Part 2: Inference from original data
# 
# This is tricky. The original data is referred to in [CASA Docs https://casadocs.readthedocs.io/en/stable/notebooks/memo-series.html#Flux-Calibrator-Models]. However exactly how this is done for each source is a bit opaque.
# 
# In this part, we use the actual measurement data and estimates of likelihood to infer a source model with quantified uncertainty. We'll use the source J1939-6342, also known as 1934-638 in the B1950. The parameters of this source are determined (apparently) be references 1,3,4,5,6,8.
# 
# Standards are: 
# * (1) Perley-Butler 2010, 
# * (2) Perley-Butler 2013, 
# * (3) Perley-Taylor 99,   
# * (4) Perley-Taylor 95
# * (5) Perley 90, 
# * (6) Baars,
# * (7) Scaife-Heald 2012, 
# * (8) Stevens-Reynolds 2016 [https://ui.adsabs.harvard.edu/link_gateway/2016ApJ...821...61P/doi:10.3847/0004-637X/821/1/61]
# 
# The real source of data for J1939-6342 appears to be Reynolds [Reynolds, J. E. A revised flux scale for the AT Compact Array. AT Memo 39.3/040, 1994. https://archive-gw-1.kat.ac.za/public/meerkat/A-Revised-Flux-Scale-for-the-AT-Compact-Array.pdf] in the range 408MHz - 8.6GHz. And from [Partridge, B., et al. "Absolute calibration of the radio astronomy flux density scale at 22 to 43 GHz using Planck." The Astrophysical Journal 821.1 (2016): 61.] in the higher frequency ranges.
# 

# In[3]:


import numpy as np
import matplotlib.pyplot as plt
import json
import os
import arviz as az
from specfit import inference
from specfit import posterior_helper
import h5py
import specfit


# The following coefficients are from J.E. Reynolds.

# In[2]:


src = { 'name': 'J1939-6342',
        'nu0': 1e6,
        'a': [-30.7667, 26.4908, -7.0977, 0.605334]}
name = src['name']
nu0 = src['nu0']
a_array = src['a']


# In[3]:


def log_flux(w, logw, a):
    logS = a[0]
    for i in range(1,len(a)):
        logS += a[i]*logw**i
        
    return logS

def flux(nu, a, nu0):
    w = nu/nu0
    logw = np.log(w)
    
    logS = log_flux(w, logw, a)
    return np.exp(logS)


# In[4]:


fig, ax0 = plt.subplots(nrows=1, sharex=True)
nu = np.linspace(0.5e9, 10e9, 100)

f = flux(nu, src['a'],src['nu0'])
ax0.loglog(nu/1e9, f, '-', label=src['name'])
ax0.grid(True)
# ax0.yaxis.set_major_formatter(mpl.ticker.ScalarFormatter())
#ax0.set_xscale("log", nonpositive='clip')
#ax0.set_yscale("log", nonpositive='clip')
ax0.set_xlabel("Frequency (GHz)")
ax0.set_ylabel("Flux (Jy)")
plt.title(f"Polynomial Model Source Spectrum")
ax0.legend();
plt.tight_layout()
plt.savefig('source_original_spectrum.pdf')


# ## Original Data
# 
# We'll reconstruct the fit from the original data of Reynolds (1994). A discussion of this data is avaialable [here](https://www.narrabri.atnf.csiro.au/observing/users_guide/html_old_20090512/Flux_Density_Measurements.html)

# In[5]:


import pymc as pm
import pytensor.tensor as tt


# In[6]:


default_percent_uncertainty = 5.0

# AT Compact Array using the calibrator 3C286 as the reference. 
# The total flux was estimated by using only the shortest interferometer spacings or
# by extrapolating the flux density to zero spacing (eg McConnell & McKay 1994). 
d_atca = default_percent_uncertainty / 100

# Single dish using parkes and Hydra A (occasionally Virgo A as reference)
d_parkes = default_percent_uncertainty / 100 
d_jer = default_percent_uncertainty / 100 ## J.E. Reynolds 1994 Tidbinbilla single dish using parkes and Hydra A (occasionally Virgo A as reference)

# The Molonglo Reference Catalogue (MRC) at 408MHz is on the scale of Wyllie (1969) 
d_mrc = default_percent_uncertainty / 100 

# MOST (Campbell-Wilson & Hunstead 1994) The MOST flux scale at 843MHz is essentially an 
# interpolation between the Wyllie 408MHz scale and the scale of the Parkes Catalogue.
# See Hunstead (1991) for further details.
d_most = default_percent_uncertainty / 100 

orig_measured_data = np.array(
                [[0.408, 6.24, d_mrc],
                [0.843, 13.65, d_most],
                [1.380, 14.96, d_atca],
                [1.413, 14.87, d_parkes],
                [1.612, 14.47, d_parkes],
                [1.660, 14.06, d_parkes],
                [1.665, 14.21, d_jer],
                [2.295, 11.95, d_jer],
                [2.378, 11.75, d_atca],
                [4.800, 5.81, d_atca],
                [4.800, 5.76, d_atca],
                [4.835, 5.72, d_atca],
                [4.850, 5.74, d_parkes],
                [8.415, 2.99, d_atca],
                [8.420, 2.97, d_jer],
                [8.640, 2.81, d_atca],
                [8.640, 2.81, d_atca]])

orig_measured_data = orig_measured_data.T


# In[46]:


nu = orig_measured_data[0]*1e9
measured_data = orig_measured_data[1]
percent = orig_measured_data[2]
sigma = measured_data*percent

new_d = np.array([nu/1e9, measured_data, sigma]).T
#print(f"Original Data:\n{np.array2string(new_d, separator=', ', precision=4)}")

min_freq = nu[0]
max_freq = nu[-1]

#nu, measured_data


# In[8]:


fig, ax0 = plt.subplots(nrows=1, sharex=True, figsize=(8,6))
ax0.set_xscale("log", nonpositive='clip')
ax0.set_yscale("log", nonpositive='clip')
ax0.errorbar(nu/1e9, measured_data, yerr=sigma, fmt='.')
ax0.grid(True)
ax0.set_title('Flux Measurements J1939-6342')
ax0.set_xlabel("Frequency (GHz)")
ax0.set_ylabel("Flux (Jy)")
fig.tight_layout()
plt.savefig('j1939_measurements.pdf')


# ## Choosing priors
# 
# These need to be chosen carefully. I0 is the value of the flux at $\nu = \nu_0$, and in this case we choose $\nu_0$ to be one of the data points close to the usual choice.

# In[9]:


nu0 = 1.0e9 # nu[3];


order=4
with pm.Model() as model:
    
    # Choose priors
    a_1  = [ pm.Normal(f"a[{i}]", mu=0, sigma=2.5) for i in range(order) ]
    print(a_1)
    brightness_1 = flux(nu, a_1, nu0)

    # likelihood_1 = pm.Normal("likelihood", mu=brightness_1, sigma=sigma, observed=measured_data)
    likelihood_1 = pm.StudentT("likelihood", nu=5, mu=brightness_1, sigma=sigma, observed=measured_data)


# In[10]:


with model:
    prior_checks = pm.sample_prior_predictive(samples=50, return_inferencedata=False)

print("test")
print(prior_checks.keys())
_, ax = plt.subplots()

ax.set_xscale("log", nonpositive='clip')
ax.set_yscale("log", nonpositive='clip')

for a0, a1, a2, a3 in zip( prior_checks["a[0]"],
                   prior_checks["a[1]"],
                   prior_checks["a[2]"],
                   prior_checks["a[3]"]
                             ):
    b = flux(nu, [a0, a1, a2, a3], nu0)
    ax.plot(nu/1e9, b, c="k", alpha=0.4)
ax.plot(nu/1e9, measured_data, c="b", alpha=1, label="J1939-6342")

ax.set_xlabel("Frequency (GHz)")
ax.set_ylabel("Flux (Jy)")
ax.set_title("Prior predictive checks");
ax.grid(True)
ax.legend()
plt.tight_layout()
plt.savefig("original_prior_predictive.pdf")


# Now sample from the model, producing a chain...

# In[11]:


idata_j1939 = inference.run_or_load(model, fname = "idata_j1939.nc",
                         n_samples = 5000, n_tune=3000, n_chains=4)


# ## Results
# 
# 

# In[12]:


stats = pm.summary(idata_j1939)
print(stats.to_string())


# In[13]:


az.plot_pair(
        idata_j1939,
        var_names=['a['],
        kind="hexbin",
        filter_vars="like",
        marginals=True,
        figsize=(12, 12),
    );
plt.savefig('posterior_pairs_J1939.pdf')


# The covariance of the parameters is clearly significant. We can estimate this numerically.

# In[14]:


a_cov, a_corr, names = posterior_helper.chain_covariance(idata_j1939.posterior)
np.set_printoptions(precision=4, suppress=False)
a_cov
stats, names = posterior_helper.get_stats(idata_j1939.posterior)

with h5py.File('calibrator_catalogue.hdf5', 'a') as f:
    grp = f.require_group("J1939-6342")
    ds_mean = grp.create_dataset("mean", data=stats[0])
    ds_sdev = grp.create_dataset("sdev", data=stats[1])
    ds_cov = grp.create_dataset("cov", data=a_cov)
    ds_cor = grp.create_dataset("corr", data=a_corr)

# In[15]:

ofile = open("j1939.tex", 'w')

posterior_helper.full_column(outfile=ofile, all_names=["J1939-6342"], idata=idata_j1939, freq=nu)  # outfile, all_names, idata, freq


# ### Posterior Predictive Sampling
# 
# Sanity check to see that we are producing reasonable spectra.

# In[18]:


RANDOM_SEED=123
with model:
    ppc = pm.sample_posterior_predictive(
        idata_j1939, random_seed=RANDOM_SEED, return_inferencedata=False
    )


# In[19]:


#ppc
#az.plot_ppc(az.from_pymc3(posterior_predictive=ppc, model=model));


# In[20]:


_, ax = plt.subplots()
ax.plot(nu, measured_data, "o", ms=4, alpha=0.4, label="Data")
az.plot_hdi(
    nu,
    ppc["likelihood"],
    ax=ax,
    fill_kwargs={"alpha": 0.8, "color": "#a1dab4", "label": "Outcome 94% HDI"},
)

ax.set_xlabel("Frequency (GHz)")
ax.set_ylabel("Flux (Jy)")
ax.set_title("Posterior predictive checks")
ax.legend(ncol=2, fontsize=10);
plt.savefig('posterior_predictive_j1939.pdf')


# ## Validation against usual fit
# 
# Now have a look at the new spectral fit, taken by using the mean of the posterior parameters, against the polynomial of Reynolds (1994).

# In[21]:


a = [ idata_j1939.posterior.mean().get(f"a[{i}]").values.tolist() for i in range(order)]


# In[22]:


print(f"The mean value nu0 = {nu0} and a={a}, i0={10**a[0]}")


# Now form the spectrum from the mean posterior values and compare it to the Reynolds model (using the Perley butler source spectral characteristics)

# In[23]:


nu_new = np.linspace(0.4e9, 10e9, 1000)
b_new = flux(nu_new, a, nu0)
b_pb = flux(nu_new, src['a'],src['nu0'])


# In[24]:


_, ax = plt.subplots()

ax.set_xscale("log", nonpositive='clip')
ax.set_yscale("log", nonpositive='clip')
ax.plot(nu_new, b_pb, c="k", alpha=0.4, label="Reynolds (1994)")
ax.plot(nu_new, b_new, '-.', c="g", alpha=1, label="This work")
ax.plot(nu_new, np.abs(b_new-b_pb), c="r", alpha=1, label='difference $\\Delta S$')
ax.plot(nu, measured_data, 'o', c="b", alpha=1, label="Measurements")

ax.legend()
ax.set_xlabel("Frequency (GHz)")
ax.set_ylabel("Flux (Jy)")
ax.set_ylim((0.001, 20))
ax.grid(True)
ax.set_title("Comparison of Models");
plt.savefig('model_comparison.pdf')


# In[25]:


ds =(np.abs(b_new - b_pb))
np.percentile(ds, [0.05, 0.5, 0.95, 0.99, 1])


# In[26]:


import matplotlib as mpl

def mean_val(key):
    return idata_j1939.posterior[key].mean().values.tolist()

a_mean = [mean_val(f"a[{i}]") for i in range(order)]


fig, ax = plt.subplots()


ax.set_xscale("log", nonpositive='clip')
ax.set_yscale("log")
for n in np.geomspace(min_freq, max_freq, 50):
    for i in range(100):
        a = posterior_helper.get_random_sample(idata_j1939)
        b_test = flux(n, a, nu0)

        if False:
            b_ref = flux(n, a_mean, nu0)
            ax.plot(n/1e9, 1000*(b_ref-b_test), '.', c="k", alpha=0.05)
        else:
            ax.plot(n/1e9, b_test, '.', c="k", alpha=0.05)

ax.set_xlabel("Frequency (GHz)")
ax.set_ylabel("Flux (Jy))")
ax.set_title("Posterior Spectrum PDF (J1939-6342)")
ax.grid(True);
fig.tight_layout()
plt.savefig('posterior_spectrum_j1939.pdf')


# In[27]:


## Now sample from a multivariate gaussian of the polynomial model
order=4
mean = [ idata_j1939.posterior.mean().get(f"a[{i}]").values.tolist() for i in range(order)]
a_cov, a_corr, names = posterior_helper.chain_covariance(idata_j1939.posterior)
print(mean)
print(a_cov)
samples_cov = np.random.multivariate_normal(mean, a_cov, 100)

diag_cov = np.zeros_like(a_cov)
diag_cov[np.diag_indices(a_cov.shape[0])] = np.diag(a_cov)
print(diag_cov)
samples_diag = np.random.multivariate_normal(mean, diag_cov, 100)
fig, ax = plt.subplots(2, sharex=True, sharey=True, figsize=(7,7))
ax0, ax1 = ax

ax0.set_xscale("log", nonpositive='clip')
ax0.set_yscale("log")

fifth = []
fiftieth = []
ninetyfifth = []

nu = np.geomspace(min_freq, max_freq, 50)

for n in nu:
    fluxes = np.array([flux(n, a, nu0) for a in samples_cov])
    for s in fluxes:
        ax0.plot(n/1e9, s, '.', c="k", alpha=0.05)
ax0.set_title("Posterior with covariance")
ax0.grid(True);
ax0.set_ylabel("Flux (Jy))")

for n in nu:
    fluxes = np.array([flux(n, a, nu0) for a in samples_diag])
    for s in fluxes:
        ax1.plot(n/1e9, s, '.', c="k", alpha=0.05)
ax1.set_title("Posterior without covariance")
ax1.grid(True);
ax1.set_xlabel("Frequency (GHz)")
ax1.set_ylabel("Flux (Jy))")
        
#    b_test = flux(n, a, nu0)

#        if False:
#            b_ref = flux(n, a_mean, nu0)
#            ax.plot(n/1e9, 1000*(b_ref-b_test), '.', c="k", alpha=0.05)
#        else:
#            ax.plot(n/1e9, b_test, '.', c="k", alpha=0.1)

#fig.set_title("Posterior Spectrum PDF (J1939-6342)")
fig.tight_layout()
plt.savefig('posterior_spectrum_model_j1939.pdf')


# ## Discussion
# 
# The new polynomial fit is fairly accurate, within 50 $\mu$Jy. Of more importance is that the uncertainty in the polynomial coefficients can be quantified, and mapped through to uncertainty in the spectrum. This allows the parameters of the model to be used as priors for a Bayesian calibration.

# ## Northern Sky Calibrators
# 
# Table 9
# Adopted Spectral Flux Densities of Steady Sources
# 
# Notes. The derived flux density values, based on the Mars emission model for frequencies between 1 and 50 GHz. The quoted errors are derived from the dispersion of the values over the various sessions. The values for 3C123, 3C286, and 3C295 at 327.5 MHz are derived from their ratios to 3C196, whose flux density is taken to be 46.8 Jy (Scaife & Heald 2012).
# 

# In[28]:


# Freq., 3C123, , 3C196, , 3C286, , 3C295, , N_obs
# (GHz), S(Jy), sigma(Jy), S(Jy), sigma(Jy), S(Jy), sigma(Jy), S(Jy), sigma(Jy)

csv_data = '''0.3275, 145.0, 4.3, 46.8, 1.4, 26.1, 0.8, 60.8, 1.8, 14
1.015, 66.2, 4.3, 20.1, 4.8, 18.4, 4.3, 30.8, 7.3, 1
1.275, 46.6, 3.2, 13.3, 2.0, 13.8, 2.0, 21.5, 3.0, 1
1.465, 47.8, 0.5, 14.1, 0.2, 15.0, 0.2, 22.2, 0.5, 4
1.865, 38.7, 0.6, 11.3, 0.2, 13.2, 0.2, 17.9, 0.3, 2
2.565, 28.9, 0.3, 8.16, 0.1, 10.9, 0.2, 12.8, 0.2, 2
3.565, 21.4, 0.8, 6.22, 0.2, 9.5, 0.1, 9.62, 0.2, 3
4.535, 16.9, 0.2, 4.55, 0.06, 7.68, 0.1, 6.96, 0.09, 7
4.835, 16.0, 0.2, 4.22, 0.1, 7.33, 0.2, 6.45, 0.15, 1
4.885, 15.88, 0.1, 4.189, 0.025, 7.297, 0.046, 6.37, 0.04, 11
6.135, 12.81, 0.15, 3.318, 0.05, 6.49, 0.15, 4.99, 0.05, 2
6.885, 11.20, 0.14, 2.85, 0.05, 5.75, 0.05, 4.21, 0.05, 3
7.465, 11.01, 0.2, 2.79, 0.05, 5.70, 0.10, 4.13, 0.07, 1
8.435, 9.20, 0.04, 2.294, 0.010, 5.059, 0.021, 3.319, 0.014, 11
8.485, 9.10, 0.15, 2.275, 0.03, 5.045, 0.07, 3.295, 0.05, 1
8.735, 8.86, 0.05, 2.202, 0.011, 4.930, 0.024, 3.173, 0.016, 10
11.06, 6.73, 0.15, 1.64, 0.03, 4.053, 0.08, 2.204, 0.05, 1
12.890, , , 1.388, 0.025, 3.662, 0.070, 1.904, 0.04, 1
14.635, 5.34, 0.05, 1.255, 0.020, 3.509, 0.040, 1.694, 0.04, 1
14.715, 5.02, 0.05, 1.206, 0.020, 3.375, 0.040, 1.630, 0.03, 1
14.915, 5.132, 0.025, 1.207, 0.004, 3.399, 0.016, 1.626, 0.008, 7
14.965, 5.092, 0.028, 1.198, 0.007, 3.387, 0.015, 1.617, 0.007, 11
17.422, 4.272, 0.07, 0.988, 0.02, 2.980, 0.04, 1.311, 0.025, 1
18.230, , , 0.932, 0.020, 2.860, 0.045, 1.222, 0.05, 1
18.485, 4.090, 0.055, 0.947, 0.015, 2.925, 0.045, 1.256, 0.020, 1
18.585, 3.934, 0.055, 0.926, 0.015, 2.880, 0.04, 1.221, 0.015, 1
20.485, 3.586, 0.055, 0.820, 0.010, 2.731, 0.05, 1.089, 0.015, 1
22.460, 3.297, 0.022, 0.745, 0.003, 2.505, 0.016, 0.952, 0.005, 13
22.835, 3.334, 0.06, 0.760, 0.010, 2.562, 0.05, 0.967, 0.015, 1
24.450, 2.867, 0.03, 0.657, 0.017, 2.387, 0.03, 0.861, 0.020, 2
25.836, 2.697, 0.06, 0.620, 0.017, 2.181, 0.06, 0.770, 0.02, 1
26.485, 2.716, 0.05, 0.607, 0.017, 2.247, 0.05, 0.779, 0.020, 1
28.450, 2.436, 0.06, 0.568, 0.015, 2.079, 0.05, 0.689, 0.020, 2
29.735, 2.453, 0.05, 0.529, 0.015, 2.011, 0.05, 0.653, 0.020, 1
36.435, 1.841, 0.17, 0.408, 0.005, 1.684, 0.02, 0.484, 0.015, 3
43.065, , , 0.367, 0.015, 1.658, 0.08, 0.442, 0.020, 1
43.340, 1.421, 0.055, 0.342, 0.005, 1.543, 0.024, 0.398, 0.006, 13
48.350, 1.269, 0.12, 0.289, 0.005, 1.449, 0.04, 0.359, 0.013, 4
48.565, , , 0.272, 0.015, 1.465, 0.1, 0.325, 0.025, 1
'''

'''
# In[29]:


import csv
import io
import numpy as np
buff = io.StringIO(csv_data)

def parse(x):
    try:
        return float(x)
    except:
        return None
    
reader = csv.reader(buff)
lines = []
for row in reader:
    
    line = [parse(x) for x in row]
    lines.append(line)


def cleanup(line, sigma, freq):
    good = np.where(np.isnan(line) == False)
    return line[good], sigma[good], freq[good]

lines = np.array(lines, dtype=np.float64).T
frequency = lines[0]*1e9
S_3C123 = lines[1]
sigma_3C123 = lines[2]

S_3C196 = lines[3]
sigma_3C196 = lines[4]

S_3C286 = lines[5]
sigma_3C286 = lines[6]

S_3C295 = lines[7]
sigma_3C295 = lines[8]

min_freq = frequency[0]
max_freq = frequency[-1]

S_3C123, sigma_3C123, f_3C123 = cleanup(S_3C123, sigma_3C123, frequency)


# In[30]:



def dataplot(name, freq, mu, sigma):
    fig, ax = plt.subplots()

    ax.set_xscale("log", nonpositive='clip')
    ax.set_yscale("log", nonpositive='clip')
    ax.errorbar(freq/1e9, mu, yerr=sigma, fmt='.', label="Perley & Butler")

    ax.set_xlabel("Frequency (GHz)")
    ax.set_ylabel("Flux (Jy)")
    ax.grid(True)
    ax.set_title(f"Flux Measurements: {name}");
    fig.tight_layout()
    plt.savefig(f"source_{name}_spectrum.pdf")
    
def spectral_inference(name, freq, mu, sigma, order, nu0):
    dataplot(name, freq, mu, sigma)
    with pm.Model() as _model:
        _a  = [ pm.Normal(f"a[{i}]", mu=0, sigma=2.5) for i in  range(order) ]
        _brightness = flux(freq, _a, nu0)
        _likelihood = pm.Normal("likelihood", mu=_brightness, sigma=sigma, observed=mu)
    _idata = posterior_helper.run_or_load(_model, fname = f"idata_{name}.nc")
    
    posterior_helper.full_column(name, _idata, freq)

    
S_3C286, f_3C123


# In[31]:


nu0 = 1e9
spectral_inference("3C123", f_3C123, S_3C123, sigma_3C123, order=4, nu0=nu0)


# In[32]:


spectral_inference("3C196", frequency, S_3C196, sigma_3C196, order=4, nu0=nu0)


# In[33]:


import matplotlib.pyplot as plt

fig, ax = plt.subplots()

ax.set_xscale("log", nonpositive='clip')
ax.set_yscale("log", nonpositive='clip')
ax.errorbar(frequency/1e9, S_3C286, yerr=sigma_3C286, fmt='.', label="Perley & Butler")

#ax.legend()
ax.set_xlabel("Frequency (GHz)")
ax.set_ylabel("Flux (Jy)")
ax.grid(True)
ax.set_title("Flux Measurements 3C286");
fig.tight_layout()
plt.savefig('source_3c286_spectrum.pdf')


# In[34]:


nu0 = 1e9
order=4

    
with pm.Model() as model_3c286:
    
    a_1  = [ pm.Normal(f"a[{i}]", mu=0, sigma=2.5) for i in  range(order) ]

    brightness_1 = flux(frequency, a_1, nu0)

    likelihood_1 = pm.Normal("likelihood", mu=brightness_1, sigma=sigma_3C286, observed=S_3C286)
    
with pm.Model() as model_3c295:
    
    a_1  = [ pm.Normal(f"a[{i}]", mu=0, sigma=2.5) for i in  range(order) ]

    brightness_1 = flux(frequency, a_1, nu0)

    likelihood_1 = pm.Normal("likelihood", mu=brightness_1, sigma=sigma_3C295, observed=S_3C295)


# In[35]:


idata_3c286 = posterior_helper.run_or_load(model_3c286, fname = "idata_3c286.nc")
idata_3c295 = posterior_helper.run_or_load(model_3c295, fname = "idata_3c295.nc")


# In[36]:


stats = pm.summary(idata_3c286)
print(stats.to_string())


# In[37]:


stats = pm.summary(idata_3c295)
print(stats.to_string())


# In[38]:


az.plot_pair(
        idata_3c286,
        var_names=['a['],
        kind="hexbin",
        filter_vars="like",
        marginals=True,
        figsize=(12, 12),
    );
plt.savefig('posterior_pairs_3c286.pdf')


# In[39]:


posterior_helper.idata_2_latex(idata_3c286)


# This should be compared to the values provided by Perley & Butler (2017) of
# 
# 1.2515 ± 0.0048 	−0.4605 ± 0.0163 	−0.1715 ± 0.0208 	0.0336 ± 0.0082

# In[40]:


a_cov, a_corr, names = posterior_helper.chain_covariance(idata_3c286)
np.set_printoptions(precision=4, suppress=False)
print("Covariance")
print(a_cov)
print("Correlation")
print(a_corr)


# In[41]:

ofile = open("3C286.tex", 'w')

posterior_helper.full_column(ofile, "3C286", idata_3c286, frequency)


# In[42]:


posterior_helper.full_column(ofile, "3C295", idata_3c295, frequency)


# In[43]:


a_cov, a_corr, names = posterior_helper.chain_covariance(idata_3c286)


# In[44]:


def mean_val(key):
    return idata_3c286.posterior[key].mean().values.tolist()

a_mean = [mean_val(f"a[{i}]") for i in range(order)]


_, ax = plt.subplots()


ax.set_xscale("log", nonpositive='clip')
ax.set_yscale("log")
for n in np.geomspace(min_freq, max_freq, 50):
    for i in range(100):
        a = posterior_helper.get_random_sample(idata_3c286)
        b_test = flux(n, a, nu0)

        if False:
            b_ref = flux(n, a_mean, nu0)
            ax.plot(n/1e9, 1000*(b_ref-b_test), '.', c="k", alpha=0.05)
        else:
            ax.plot(n/1e9, b_test, '.', c="k", alpha=0.1)

ax.set_xlabel("Frequency (GHz)")
ax.set_ylabel("Flux (Jy))")
ax.set_title("Posterior Spectrum PDF (3C286)")
ax.grid(True);
plt.savefig('posterior_spectrum_3c286.pdf')


# ## 

'''




import pymc as pm
import numpy as np
import arviz as az

import specfit as sf
import matplotlib.pyplot as plt
import sympy as sp

az.style.use("arviz-darkgrid")



original_data = np.array(
    [[0.408,  6.24, 0.312 ],
     [0.843, 13.65, 0.6825],
     [1.38 , 14.96, 0.748 ],
     [1.413, 14.87, 0.7435],
     [1.612, 14.47, 0.7235],
     [1.66 , 14.06, 0.703 ],
     [1.665, 14.21, 0.7105],
     [2.295, 11.95, 0.5975],
     [2.378, 11.75, 0.5875],
     [4.8  ,  5.81, 0.2905],
     [4.8  ,  5.76, 0.288 ],
     [4.835,  5.72, 0.286 ],
     [4.85 ,  5.74, 0.287 ],
     [8.415,  2.99, 0.1495],
     [8.42 ,  2.97, 0.1485],
     [8.64 ,  2.81, 0.1405]])

freq_ghz, S, delta_S = original_data.T
freq = freq_ghz*1e9
nu0=1.0e9


name = "j1939-6342"
sigma = delta_S

names, stats, a_cov, a_corr, idata, model, knot_list = \
        sf.spline_inference(name, freq, S, sigma, nu0)

# spline_model = sf.get_spline_model(name, freq, mu, sigma, order, nu0=nu0)
# 
# with spline_model:
#     idata = pm.sample_prior_predictive()
#     idata.extend(pm.sample(draws=1000, tune=3000, random_seed=RANDOM_SEED, chains=4))
#     pm.sample_posterior_predictive(idata, extend_inferencedata=True)

print(idata.keys())
print(az.summary(idata, var_names=[ "w"]))

az.plot_trace(idata, var_names=[ "w"]);
plt.show()
plt.savefig(f"spline_trace.pdf")


## Pair Plots

plt.clf()
az.plot_pair(
        idata,
        var_names=['w'],
        kind="hexbin",
        filter_vars="like",
        marginals=True,
        figsize=(12, 12),
    )
plt.tight_layout()
plt.savefig(f"spline_posterior_pairs.pdf")
plt.show()


#  Do some posterior sampling
sf.plot_spline_design(freq, S, nu0, knot_list)

sf.plot_spline(idata, freq, S, nu0, knot_list)





import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import specfit as sf
import pymc as pm
import arviz as az
import h5py
import logging

logger = logging.getLogger('specfit')
logger.setLevel(logging.INFO)

# create console handler and set level to debug
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)

# create formatter
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

# add formatter to ch
ch.setFormatter(formatter)

# add ch to logger
logger.addHandler(ch)

NU_0 = 1.0e9


def cleanup(_S, _sigma, _freq):
    S = np.array(_S)
    sigma = np.array(_sigma)
    freq = np.array(_freq)
    good = np.where(np.isnan(S) == False)
    return S[good], sigma[good], freq[good]


def process_bic(outfile, official_name, key, _S, _sigma, _frequency, order):
    print(f"Processing {key}")
    nu0 = NU_0
    S, sigma, freq = cleanup(_S, _sigma, _frequency)

    bic_data = sf.marginal_likelihood(key, freq, S, sigma, nu0=nu0)

    bmax = np.max(bic_data[:,1])

    bics = np.exp(bic_data[:,1] - bmax)
    orders = bic_data[:,0]

    print(f"{official_name} & {freq[0]/1e9 :4.2f}-{freq[-1]/1e9 :4.2f} &  {order}", file=outfile, end='')
    for _o, _b in zip(orders, bics):
        if _b == 1:
            print(f" &  \\bf{{{_b :5.4f}}} ", file=outfile, end='')
        else:
            print(f" &  {_b :5.4f} ", file=outfile, end='')
    print('\\\\', file=outfile, flush=True)


def process(outfile, official_names, key, _S, _sigma, _frequency, order):

    print(f"%%%%% Processing {key} order={order}", flush=True)
    print(f"%%%%% Processing {key} order={order}", file=outfile, flush=True)
    nu0 = NU_0
    S, sigma, freq = cleanup(_S, _sigma, _frequency)
    print(f" S = {S}")
    print(f" sigma = {sigma}")
    print(f" freq = {freq}")
    # bic = sf.data_inference(key, freq, S, sigma, nu0=nu0, order=order)

    names, stats, a_cov, a_corr, idata, mcmc_model = sf.data_inference(key, freq, S, sigma, order=order, nu0=nu0)

    with h5py.File('calibrator_catalogue.hdf5', 'a') as f:
        grp = f.require_group(key)
        ds_mean = grp.require_dataset("mean", shape=stats[0].shape, dtype=stats[0].dtype, data=stats[0])
        ds_sdev = grp.require_dataset("sdev", shape=stats[1].shape, dtype=stats[1].dtype, data=stats[1])
        ds_cov = grp.require_dataset("cov", shape=a_cov.shape, dtype=a_cov.dtype, data=a_cov)
        ds_cor = grp.require_dataset("corr", shape=a_corr.shape, dtype=a_corr.dtype, data=a_corr)


    sf.full_column(outfile, official_names, idata.posterior, np.array(freq))
    #print(pm.summary(idata))
    fig, ax = sf.dataplot(plt, official_names[0], freq, S, sigma)
    a, da = stats
    nu = np.geomspace(freq[0], freq[-1], 100)
    ax.plot(nu/1e9, sf.flux(nu, a, nu0), label='Polynomial model')

    # We wan't to do in-sample predictions at the new range of frequencies (nu)

    # ppc = sf.posterior_predictive_sampling(idata, mcmc_model, num_pp_samples=2000)
    with mcmc_model:
        # pm.set_data({'frequencies': nu})
        ppc = pm.sample_posterior_predictive(idata,
                                             var_names=['likelihood', 'frequencies']).posterior_predictive

    print(ppc.keys())
    posterior_pred = ppc['likelihood'].mean(axis=0)   # Average over the chains
    print(f"posterior pred: {posterior_pred.shape}")

    posterior_freq = ppc['frequencies'].mean(axis=0).mean(axis=0)    # Average over the chains, then over the samples
    print(f"posterior freq: {posterior_freq.shape}")

    y_mean = posterior_pred.mean(axis=0)    # Average over samples
    y_std = posterior_pred.std(axis=0)      # Std over samples

    print(f"y_mean: {y_mean.shape}")
    print(f"y_std: {y_std.shape}")

    # ax.plot(posterior_freq/1e9, y_mean, label='Posterior mean')
    # ax.plot(nu/1e9, sf.flux(nu, y_mean, nu0), label='Posterior flux mean')
    ax.fill_between(posterior_freq/1e9, y_mean - 3*y_std, y_mean + 3*y_std, alpha=0.33, label='Uncertainty Interval ($\mu\pm3\sigma$)')

    ax.legend()
    #plt.show()
    fig.savefig(f"./output/{key}_data.pdf")
    plt.close(fig)

    fig, ax = sf.posterior_plot(plt, official_names[0], freq, idata, nu0)
    fig.savefig(f"./output/{key}_pdf.pdf")
    plt.close(fig)

    az.plot_pair(
            idata.posterior,
            var_names=['a['],
            kind="hexbin",
            filter_vars="like",
            marginals=True,
            figsize=(12, 12),
        );
    plt.savefig(f"./output/{key}_posterior_pairs.pdf")


if __name__ == "__main__":

    if False:
        # This code is for the 2013 perley butler catalog
        data = pd.read_csv('perley_butler_ascii.txt', skiprows=5, skipfooter=1, skipinitialspace=True)

        print(data)

        frequency = data['(GHz)']*1e9

        S_3C123 = data['S(Jy)']
        sigma_3C123 = data['sigma(Jy)']

        S_3C196 = data['S(Jy).1']
        sigma_3C196 = data['sigma(Jy).1']

        S_3C286 = data['S(Jy).2']
        sigma_3C286 = data['sigma(Jy).2']

        S_3C295 = data['S(Jy).3']
        sigma_3C295 = data['sigma(Jy).3']

        N_obs = data['N_obs']

        process("3C123_2013", S_3C123, sigma_3C123)
        process("3C196_2013", S_3C196, sigma_3C196)
        process("3C286_2013", S_3C286, sigma_3C286)
        process("3C295_2013", S_3C295, sigma_3C295)


    ### Now 2017 data


    data = pd.read_csv('2016Flux-B.dat', skiprows=6, sep='\s+', skipfooter=1, skipinitialspace=True, quotechar="'", keep_default_na=True, engine='python')
    #print(data)

    frequency = data['Freq']*1e9
    err = np.array(data[r"%Err"])

    stars  = ["J0133", "For_A", "3C48", "3C123", "J0444", "3C138", "PicA", "3C144", "3C147", "3C196", "3C218", "3C274", "3C286", "3C295", "3C348", "3C353", "3C380", "3C405", "3C444", "3C461"]
    orders = {
        'J0133': 3, # J0133-3629
        'For_A': 2, # J0322-3712 Fornax A
        '3C48': 4,  # J0137+3309 3C48
        '3C123': 5, # J0437+2940 3C123
        'J0444': 3, # J0444-2809
        '3C138': 5, # J0521+1638 3C138
        'PicA': 3,  # J0519-4546 Pictor A
        '3C144': 4, # J0534+2200 3C144, Taurus A, Crab
        '3C147': 6, # J0542+4951 3C147
        '3C196': 5, # J0813+4813 3C196
        '3C218': 5, # J0918-1205 3C218, Hydra A
        '3C274': 3, # J1230+1223 3C274, Virgo A, M87
        '3C286': 4, # J1331+3030 3C286
        '3C295': 5, # J1411+5212 3C295
        '3C348': 3, # J1651+0459 3C348, Hercules A
        '3C353': 4, # J1720-0058 3C353
        '3C380': 6, # J1829+4844 3C380
        '3C405': 5, # J1959+4044 3C405, Cygnus A
        '3C444': 4, # J2214-1701 3C444
        '3C461': 4  # J2323+5848 3C461, Cassiopeia A
    }

    full_names = {
        'J0133': ['J0133-3629'],
        'For_A': ['J0322-3712', 'Fornax A'],
        '3C48':  ['J0137+3309', '3C48'],
        '3C123': ['J0437+2940', '3C123'],
        'J0444': ['J0444-2809'],
        '3C138': ['J0521+1638', '3C138'],
        'PicA':  ['J0519-4546', 'Pictor A'],
        '3C144': ['J0534+2200', '3C144', 'Taurus A', 'Crab'],
        '3C147': ['J0542+4951', '3C147'],
        '3C196': ['J0813+4813', '3C196'],
        '3C218': ['J0918-1205', '3C218', 'Hydra A'],
        '3C274': ['J1230+1223', '3C274', 'Virgo A', 'M87'],
        '3C286': ['J1331+3030', '3C286'],
        '3C295': ['J1411+5212', '3C295'],
        '3C348': ['J1651+0459', '3C348', 'Hercules A'],
        '3C353': ['J1720-0058', '3C353'],
        '3C380': ['J1829+4844', '3C380'],
        '3C405': ['J1959+4044', '3C405', 'Cygnus A'],
        '3C444': ['J2214-1701', '3C444'],
        '3C461': ['J2323+5848', '3C461', 'Cassiopeia A']
    }

    do_marginal_likelihood = False

    if do_marginal_likelihood:
        with open('perley_butler_bic.tex', 'w') as outfile:
            print("""
        \\begin{table}
            \\centering
            \\caption{SMC estimates of the relative marginal likelihood $B_n$, of the model vs order $n$ for various sources.}
            \\label{tab:marginal_likelihood}
        \\begin{tabular}{r|rrrrrrr}
            \\hline
            Source & $\nu$ (Ghz) & order &  $B_2$ & $B_3$ & $B_4$ & $B_5$ & $B_6$ \\\\
            \\hline
        """, file=outfile)
            for calibrator in stars:
                official_name = full_names[calibrator][0]
                mu = data[calibrator]
                sigma = mu*(err / 100)
                process_bic(outfile, official_name, calibrator, mu, sigma, frequency, orders[calibrator])
            print("""
        \\end{tabular}
        \\end{table}
        """, file=outfile)

    N = len(stars)
    split = N//2

    with open('perley_butler_2017_1.tex', 'w') as outfile:
        for calibrator in stars[0:split]:
            official_name = full_names[calibrator][0]
            mu = data[calibrator]
            sigma = mu*(err / 100)
            process(outfile, official_name, calibrator, mu, sigma, frequency, orders[calibrator])
    with open('perley_butler_2017_2.tex', 'w') as outfile:
        for calibrator in stars[split:]:
            official_name = full_names[calibrator][0]
            mu = data[calibrator]
            sigma = mu*(err / 100)
            process(outfile, official_name, calibrator, mu, sigma, frequency, orders[calibrator])



import json
import matplotlib
matplotlib.use('PDF')

import concurrent.futures

import numpy as np
import pymc as pm
import pytensor.tensor as tt

import matplotlib.pyplot as plt
import arviz as az
import specfit as sf



# helper function
def cp_design_mat(x, cps, module=tt):
    return (0.5 * (1.0 + module.sign(module.tile(x[:, None], (1, cps.shape[0])) - cps)))

def get_spline_model(name, freq, S, sigma, nu0, order=1):

    norm_f = freq/nu0
    x_data = tt.log(norm_f)

    y_err = np.array(sigma)

    if (order == 0):
        with pm.Model() as _model:

            k = pm.Normal("k", mu=0, sigma=1)  # initial slope
            m = pm.Normal("m", mu=0, sigma=1)  # offset

            logS = k*x_data + m
            D = pm.StudentT("likelihood", nu=5, mu=tt.exp(logS), sigma=y_err, 
                        observed=S)
        _var_names = ["k", "m"]
    else:
        with pm.Model() as _model:

            cps = pm.Uniform("cps",lower=norm_f[0], upper=norm_f[-1], size=order)
            logcps = tt.log(cps)
            A = cp_design_mat(x_data, logcps)

            # the following are unknown parameters that you can estimate by giving 
            # them priors and then sampling with NUTS
            k = pm.Normal("k", mu=0, sigma=1)  # initial slope
            m = pm.Normal("m", mu=0, sigma=1)  # offset
            delta = pm.Normal("delta", mu=0, sigma=2, size=order) # slope parameters

            # generate the function at x values
            # y = (k + tt.dot(A, delta))*x + (m + tt.dot(A, -cps * delta))

            # logS = pm.Deterministic("logS", (k + pm.math.dot(A, delta))*x_data +
            #                       (m + pm.math.dot(A, -logcps * delta)))
            logS = (k + pm.math.dot(A, delta))*x_data + (m + pm.math.dot(A, -logcps * delta))
            D = pm.StudentT("likelihood", nu=5, mu=tt.exp(logS), sigma=y_err, 
                        observed=S)
        _var_names = ["cps", "k", "m", "delta"]

    return _model, _var_names

def plot_spline(name, freq, S, yerr, nu0, idata, n_samples=300):
    # x axis
    
    x = np.log(freq/nu0)
    x_curve = np.linspace(x[0], x[-1], 1000)
    y_scale = 1000

    with plt.rc_context({"axes.grid": True,
                         "axes.grid.which": "both",
                         "axes.formatter.min_exponent": 2}):

        fig, ax = plt.subplots(figsize=(6, 4), layout='constrained')
        ax.set_xscale("log", nonpositive='clip')
        ax.set_yscale("log", nonpositive='clip')

        ax.set_xlabel("Frequency (GHz)")
        ax.set_ylabel("Flux (mJy)")

        for i in range(n_samples):
            sample = sf.get_random_sample(idata)

            try:
                cps = sample['cps']
                k = sample['k']
                m = sample['m']
                delta = sample['delta']

                logcps = np.log(cps)

                A = cp_design_mat(x_curve, logcps, module=np)
                logS = (k + np.dot(A, delta))*x_curve + (m + np.dot(A, -logcps * delta))
            except:
                k = sample['k']
                m = sample['m']

                logS = k*x_curve + m

            ax.plot(np.exp(x_curve), np.exp(logS)*y_scale, alpha=10/n_samples, color='k', linewidth=3)
        ax.errorbar(freq/nu0, S*y_scale, yerr=yerr*y_scale, fmt='o', label="data")
        ax.set_title(name)
        # ax.legend()
        plt.savefig(f"{name}_{n_samples}.pdf")
        # plt.show()


def inner_piecewise_linear(name, freq, S, sigma, nu0, order=1, n_samples = 3000):
    print(f"piecewise_linear(name=\"{name}\")")
    spline_model, var_names = get_spline_model(name, freq, S, sigma, nu0=nu0, order=order)

        
    n_tune = n_samples
    RANDOM_SEED = 123
    n_chains = 4

    with spline_model:
        idata = pm.sample_prior_predictive()
        idata.extend(pm.sample(draws=n_samples, tune=n_tune,
                            random_seed=RANDOM_SEED,
                            chains=n_chains))
        # pm.sample_posterior_predictive(idata, extend_inferencedata=True)

    lml = sf.process_log_marginal_likelihood(spline_model, n_samples)
    
    # Compute the LOO model comparison 
    # https://www.pymc.io/projects/docs/en/latest/learn/core_notebooks/model_comparison.html
    with spline_model:
        pm.compute_log_likelihood(idata)
        loo = az.loo(idata)
        print(loo)


    summ = az.summary(idata, var_names=var_names)
    print(summ)

    plot_spline(name, freq=freq, S=S, yerr=sigma, nu0=nu0, idata=idata, n_samples=500)
    plot_spline(name, freq=freq, S=S, yerr=sigma, nu0=nu0, idata=idata, n_samples=0)
    
    ret = get_posterior_model(idata, 1000)
    print(ret)
    ret['summary'] = str(summ)
    ret['name'] = name
    ret['order'] = order
    ret['log_marginal_likelihood'] = np.mean(lml)
    # ret['loo'] = loo
    
    with open(f"{name}.json", 'w') as json_file:
        json.dump(ret, json_file, indent=4, sort_keys=True)
  
    with plt.rc_context({"axes.grid": True, "figure.constrained_layout.use": True}):
        az.plot_trace(idata, var_names=var_names, filter_vars="like")
        plt.savefig(f"{name}_trace.pdf")
        
        az.plot_pair(
                idata.posterior,
                var_names=var_names,
                kind="hexbin",
                filter_vars="like",
                marginals=True,
                figsize=(12, 12),
            );
        plt.savefig(f"{name}_posterior_pairs.pdf")

    return idata, ret

def piecewise_linear(*args, **kwargs):
    with concurrent.futures.ProcessPoolExecutor(max_workers=1) as executor:
        future = executor.submit(inner_piecewise_linear, *args, **kwargs)
        return future.result()
    

def cps_2_slope(k, cps, delta):
    n = len(cps)
    slopes = []
    slope = k
    for i in range(n):
        ds = delta[i]
        c = cps[i]
        # print(f"slope = {slope}, x < {c}")
        slopes.append(slope)
        slope = slope + ds
    
    # print(f"slope = {slope}, x > {c}")
    slopes.append(slope)
    return slopes

def get_posterior_model(idata, samples):
    ret = {}
    try:
        slope_array = []
        cps_array = []
        for i in range(samples):
            sample = sf.get_random_sample(idata)

            cps = sample['cps']
            k = sample['k']
            m = sample['m']
            delta = sample['delta']

            cps_array.append(cps)
            slope_array.append(cps_2_slope(k, cps, delta))
            
        slope_array = np.array(slope_array)
        cps_array = np.array(cps_array)
        
        print(f"slope_array {slope_array.shape}")
        print(f"cps_array {cps_array.shape}")
        
        ret['slopes'] = np.mean(slope_array, axis=0).tolist()
        ret['slopes_sigma'] = np.std(slope_array, axis=0).tolist()
        ret['change_point'] = np.mean(cps_array, axis=0).tolist()
        ret['change_point_sigma'] = np.std(cps_array, axis=0).tolist()
    except:
        pass
    return ret
    
if __name__ == "__main__":

    if False:
        # x axis
        x = np.linspace(0, 10, 100)

        # locations on axis of the changepoints
        cps = np.array([2, 4, 7])

        A = cp_design_mat(x, cps, module=np)

        # the following are unknown parameters that you can estimate by giving 
        # them priors and then sampling with NUTS
        k = 0.0  # initial slope
        m = 0.3 # offset
        delta = np.array([1, -1, -1]) # slope parameters

        cps_2_slope(k, cps, delta)
        # generate the function at x values
        y = (k + np.dot(A, delta))*x + (m + np.dot(A, -cps * delta))
        # 
        # print(x)
        # print(y)
        
        data_diff = (np.dot(A, delta))
        print((data_diff))

        plt.plot(x, y)
        plt.show()
    else:
        original_data = np.array(
            [[0.408,  6.24, 0.312 ],
            [0.843, 13.65, 0.6825],
            [1.38 , 14.96, 0.748 ],
            [1.413, 14.87, 0.7435],
            [1.612, 14.47, 0.7235],
            [1.66 , 14.06, 0.703 ],
            [1.665, 14.21, 0.7105],
            [2.295, 11.95, 0.5975],
            [2.378, 11.75, 0.5875],
            [4.8  ,  5.81, 0.2905],
            [4.8  ,  5.76, 0.288 ],
            [4.835,  5.72, 0.286 ],
            [4.85 ,  5.74, 0.287 ],
            [8.415,  2.99, 0.1495],
            [8.42 ,  2.97, 0.1485],
            [8.64 ,  2.81, 0.1405]])

        freq_ghz, S, delta_S = original_data.T
        freq = freq_ghz*1e9
        nu0=1.0e9

        name = "j1939-6342"
        sigma = delta_S
        mu=S

        idata, ret = piecewise_linear(name, freq, mu, sigma, nu0, order=2, n_samples=3000)
        print(ret)

